#include "AStar.h"

AStar::AStar(BWAPI::Game* game) {
    Broodwar = game;
}

AStar::~AStar() {
    deleteAll();
}

void AStar::deleteAll() {
    for(map<pair<int,int>,Node*>::iterator i = closed.begin(); i != closed.end(); ++i) {
        delete(i->second);
    }
    closed.clear();
    for(map<pair<int,int>,Node*>::iterator i = open_pos.begin(); i != open_pos.end(); ++i) {
        delete(i->second);
    }
    open_f.clear();
    open_pos.clear();

}

double AStar::heuristic(pair<int, int> start, pair<int, int> end) {
    return sqrt(pow(end.first - start.first ,2.0) + pow(end.second - start.second ,2.0));
}

std::vector<AStar::Node*> AStar::getNeighbors(Node* current) {
    vector<Node*> result;
    pair<int, int> next;

    for(int x = -1; x < 2; x++) {
        for(int y = -1; y < 2; y++) {
            if ((x == 0) && (y == 0)) continue;
            double cost = 1.5;
            if ((x == 0) || (y == 0)) cost = 1.0;

            next = pair<int, int>(current->pos.first + x, current->pos.second + y);

            if (closed.find(next) != closed.end()) continue;

            if (Broodwar->isWalkable(next.first, next.second) && Broodwar->isWalkable(next.first+1, next.second) && Broodwar->isWalkable(next.first-1, next.second) && Broodwar->isWalkable(next.first+1, next.second+1)
             && Broodwar->isWalkable(next.first-1, next.second-1) && Broodwar->isWalkable(next.first+1, next.second-1) && Broodwar->isWalkable(next.first-1, next.second+1) &&
            Broodwar->isWalkable(next.first, next.second+1) && Broodwar->isWalkable(next.first, next.second-1)) {
                map<pair<int,int>,Node*>::iterator it = open_pos.find(next);
                Node* n;
                
                if (it != open_pos.end()) {
                    n = it->second;
                } else {
                    n = new Node(next);  
                }
                n->cost = cost;
                result.push_back(n);
            }
        }
    } 
    return result;
}

vector<BWAPI::Position> AStar::getPath(Node* current) {
    vector<BWAPI::Position> result;
    while(current->prev != NULL) {
        result.push_back(BWAPI::Position(current->pos.first * WALK_TILE,current->pos.second * WALK_TILE));
        current = current->prev; 
    }
    deleteAll();
    std::reverse(result.begin(),result.end());  
    return result;
}

vector<BWAPI::Position> AStar::find(BWAPI::Position s, BWAPI::Position e) {   
    Node* start = new Node(pair<int,int>((int)floor(((double) s.x() / WALK_TILE) + 0.5),(int)floor(((double) s.y() / WALK_TILE) + 0.5)));
    pair<int,int> end = pair<int,int>((int)floor(((double) e.x() / WALK_TILE) + 0.5),(int)floor(((double) e.y() / WALK_TILE) + 0.5));
    
    if (!Broodwar->isWalkable(end.first, end.second)) return getPath(start); 
    
    start->g = 0;  
    start->f = heuristic(start->pos, end);
    open_f.insert(pair<double, Node*>(start->f, start));
    open_pos.insert(pair<pair<int, int>, Node*>(start->pos, start));
    
    while(open_f.size() > 0) {

        Node* current = open_f.begin()->second;

        if ((current->pos.first * WALK_TILE < (e.x() + WALK_TILE)) && (current->pos.first * WALK_TILE > (e.x() - WALK_TILE)) && (current->pos.second * WALK_TILE < (e.y() + WALK_TILE)) && (current->pos.second * WALK_TILE > (e.y() - WALK_TILE))){
            return getPath(current);
        }

        closed.insert(pair<pair<int, int>, Node*>(current->pos, current));
        open_f.erase(open_f.begin());
        open_pos.erase(open_pos.find(current->pos));
        
        vector<Node*> neighbors = getNeighbors(current);

        for(int i = 0; i < neighbors.size(); i++) { 
            double g_score = current->g + neighbors[i]->cost;

            if (g_score < neighbors[i]->g) {
                neighbors[i]->prev = current;
                neighbors[i]->g = g_score;
                double f_score = neighbors[i]->g + heuristic(neighbors[i]->pos, end);

                map<pair<int,int>,Node*>::iterator isInOpen = open_pos.find(neighbors[i]->pos);
                
                if (isInOpen == open_pos.end()) {
                    open_pos.insert(pair<pair<int, int>, Node*>(neighbors[i]->pos, neighbors[i]));
                    open_f.insert(pair<double, Node*>(f_score, neighbors[i]));
                } else {
                    multimap<double,Node*>::iterator it = open_f.find(neighbors[i]->f);
                    for(;it != open_f.upper_bound(neighbors[i]->f); it++) {
                        if (it->second->pos == neighbors[i]->pos) break;    
                    }
                    open_f.erase(it);
                    open_pos.insert(pair<pair<int, int>, Node*>(neighbors[i]->pos, neighbors[i]));
                    open_f.insert(pair<double, Node*>(f_score, neighbors[i]));
                }
                neighbors[i]->f = f_score;
            }
        }
    } 
    return getPath(start);
}