#include "ActorCritic.h"

#define OFILE 1
#define PI 3.14159265

ActorCritic::ActorCritic(int n, double g, double e) {
    critic = new MultiLayer(n+1, 1, 7, 0.009, "LP_Critic");
    actor = new MultiLayer(n+1, 2, 5, 0.09, "LP_Actor");
    gamma = g;
    exploration = e;
    if (OFILE) file.open("bwapi-data/write/LP_log.txt");
}

ActorCritic::~ActorCritic() {
    delete(critic);
    delete(actor);
    if (OFILE) file.close();
}

void ActorCritic::save() {
    critic->save();
    actor->save();
}

double getRandom() {
    double a = 0;
    double b = 0;
    while (a == 0) a = rand()/(double)RAND_MAX;
	  while (b == 0) b = rand()/(double)RAND_MAX;
    return sqrt(-2*log(a))*cos(2*PI*b);	
}

vector<double> ActorCritic::explore(vector<double> action) {
    vector<double> next;

    double angle = getRandom() * exploration;
    if (OFILE) file << "Random: " << angle << endl;
    double angleSin = sin(angle*PI/180);
    double angleCos = cos(angle*PI/180);

    next.push_back(action[0]*angleCos - action[1]*angleSin);
    next.push_back(action[0]*angleSin + action[1]*angleCos);

    return next;
}

vector<double> ActorCritic::normalizeVector(vector<double> v) {
    vector<double> norm;
    double size = sqrt(pow(v[0],2.0) + pow(v[1],2.0));
    for (int i = 0; i < v.size(); i++) norm.push_back(v[i]/size);
    return norm;
}

vector<double> ActorCritic::run(vector<double> state) {   
    vector<double> action = actor->run(state);

    if (OFILE) {
        file << "Actor-State: ";
        for (int i = 0; i< state.size(); i++) {
            file << state[i] << " ";
        }
        file << endl << "Actor-Action: " << action[0] << " " << action[1] << endl;
    }
    return action;
}

void ActorCritic::learn(double reward, vector<double> state, vector<double> next, vector<double> error, bool supervised) {
    double c_next = 0, c_reward;
    vector<double> e;

    /* compute TD-error */
    c_reward = critic->run(state).at(0);
    if (next.size() != 0)  {
        /* scout is alive */
        c_next = critic->run(next).at(0);
        e.push_back(reward + gamma * c_next - c_reward);
    } else {
        /* scout died */
        e.push_back(reward);
    }
    
    if (OFILE) {
        file << "Reward: " << reward << endl;
        file << "Error: " << error[0] << " " << error[1] << endl;
        file << "Critic-State: " << state[0] << " " << state[1] << " " << state[2] << endl;
        file << "Critic-Next: " << next[0] << " " << next[1] << " " << next[2] << endl;
        file << "Critic estimates, error: " << c_reward << " " << c_next << " " << e[0] << endl;
    }
    /* action is better */
    if (e[0] >= 0 || supervised) {
        /* actor is learning */  
        actor->learn(error, state); 
        if (OFILE) file << "Actor is learning" << endl;  
    }
    /* critic is learning */
    critic->run(state).at(0);
    critic->learn(e, state);	

    if (OFILE) file << endl;
}
