#include "ExampleAIModule.h"
#include "Scout.h"

using namespace BWAPI;

#define DMAP 0
#define DLMAP 1
#define DPATH 1

bool analyzed, analysis_just_finished;
BWTA::Region* home;
BWTA::Region* enemy_base;

Scout* scout;

vector<pair<double,Position>> tiles;

int frameCounter;
int locationDistance;

int gui;
int isMap;
int lmap;
int speed;

ifstream in;
ofstream out;

void ExampleAIModule::onStart() {
    Broodwar->enableFlag(Flag::UserInput);
    /* api */   
    frameCounter = 0, gui = 1, speed = 0, isMap = DMAP, lmap = DLMAP;
    in.open("bwapi-data/write/LP_config.txt");
    if (in != NULL) {
        in >> gui >> isMap >> speed;
        in.close();
    }

    Broodwar->setLocalSpeed(speed);
    if (!gui) Broodwar->setGUI(false);
    if (isMap) Broodwar->setLocalSpeed(200);

    for(int i = 0; i < 24; i++) {
        tiles.push_back(pair<double,Position> (0,Position(10,10)));
    }
    /* */    

    scout = new Scout(Broodwar,UnitTypes::Protoss_Probe.sightRange()); 
}
bool sort_pair( const pair<double, Position>& i, const pair<double, Position>& j ) {
    if( i.first < j.first ) return false;
    return true;
}

void ExampleAIModule::onEnd(bool isWinner) {
    out.open("bwapi-data/write/LP_config.txt");
    out << gui << " " << isMap << " " << speed;
    out.close();

    scout->save();
    delete(scout);
}

void ExampleAIModule::onFrame() {   
    if (scout->getScout() == NULL) {
        /* set scout */
        for(std::set<Unit*>::const_iterator i=Broodwar->self()->getUnits().begin();i!=Broodwar->self()->getUnits().end();i++) { 
            if ((*i)->getType().isWorker()) {
                scout->setScout(*i);
                scout->setLocation(scout->getNewLocation());               
                locationDistance = Scout::pathSize(scout->getScout()->getPosition(),scout->getLocation(),scout->path);
            }
        } 
        if ((Broodwar->getFrameCount()%30 == 0) && (scout->getScout() == NULL)) {
            scout->res << 0;
            Broodwar->restartGame();
        }
    } else {
        /* draw */
        if (DPATH) {
            if (isMap) {
                int color = 181;
                for(int i = 0; i < tiles.size(); i++) {
                    if (i > 2 && i < tiles.size()-1 && (i%3 == 1) && tiles[i].first == tiles[i+1].first) color -=1;
                    Broodwar->drawBoxMap(tiles[i].second.x() - TILE_SIZE/2, tiles[i].second.y()  - TILE_SIZE/2, tiles[i].second.x()  + TILE_SIZE/2, tiles[i].second.y()  + TILE_SIZE/2, Color(color), true);
                    if (lmap) Broodwar->drawTextMap(tiles[i].second.x(),tiles[i].second.y(),"%d",(int)(i),Colors::Red);
                    else {
                        if (tiles[i].first >= 0) {
                            Broodwar->drawTextMap(tiles[i].second.x(),tiles[i].second.y(),"%d", (int) 1 ,Colors::Red);
                        } else {
                            Broodwar->drawTextMap(tiles[i].second.x(),tiles[i].second.y(),"%d", (int) 0, Colors::Red);
                        }
                    }
                    if (i%3 == 1) color += 1;
                }
            }

            /* path */
            Broodwar->drawLineMap(scout->getLocation().x(), scout->getLocation().y(), scout->getScout()->getPosition().x(),scout->getScout()->getPosition().y(),Colors::Green);
            Broodwar->drawCircleMap(scout->getLocation().x(), scout->getLocation().y(),2,Colors::Purple,true);
            for(int i = 0; i < scout->path.size(); i++) {
                if (i == 0)Broodwar->drawCircleMap(scout->path[i].x(), scout->path[i].y(),2,Colors::Blue,true);
                if (i == 1)Broodwar->drawCircleMap(scout->path[i].x(), scout->path[i].y(),2,Colors::Red,true);
                else if (i == scout->path.size()-1)Broodwar->drawCircleMap(scout->path[i].x(), scout->path[i].y(),2,Colors::Yellow,true);
                else Broodwar->drawCircleMap(scout->path[i].x(), scout->path[i].y(),2,Colors::Orange,true);
            } 

            if (scout->drawVector.size() == 4) {
                Broodwar->drawLineMap(scout->getScout()->getPosition().x()+ scout->drawVector[2] * TILE_SIZE, scout->getScout()->getPosition().y() + scout->drawVector[3] * TILE_SIZE, scout->getScout()->getPosition().x(),scout->getScout()->getPosition().y(),Colors::Blue);           
                Broodwar->drawLineMap(scout->getScout()->getPosition().x()+ scout->drawVector[0] * TILE_SIZE, scout->getScout()->getPosition().y() + scout->drawVector[1] * TILE_SIZE, scout->getScout()->getPosition().x(),scout->getScout()->getPosition().y(),Colors::Red);
            } 
        }

        if (Broodwar->getFrameCount()%8 == 0) {
            /* learn */
            scout->learn();

            /* restart game */
            if (scout->isOutOfMap()) {
                Broodwar->restartGame();
            }
  
            /* scout reached location */
            if (scout->reachedLocation(frameCounter)) {
                Broodwar->restartGame();
                scout->setLocation(scout->getNewLocation()); 
                frameCounter = 0;     

            } 

            frameCounter += 8;
            locationDistance = Scout::pathSize(scout->getScout()->getPosition(),scout->getLocation() ,scout->path);;
            scout->ac->file << "Frame: " << frameCounter << endl;
            
            /* generate critic estimates */
            if (isMap) {
                vector<double> state;
                double c;

                Position s = Position(scout->getScout()->getPosition().x(),scout->getScout()->getPosition().y());
                double k = 0;
                vector<double> state_old = scout->getOtherState(s, k);
                double c_old = scout->ac->critic->run(state_old).at(0);
                
                int index = 0;
                for(int x = -2; x < 3; x++) {
                    for(int y = -2; y < 3; y++) {
                        if (y == 0 && x == 0) continue;
                        s = Position(scout->getScout()->getPosition().x() + x*TILE_SIZE, scout->getScout()->getPosition().y() + y*TILE_SIZE);
                        
                        double size;
                        state = scout->getOtherState(s,size);
                        c = scout->ac->critic->run(state).at(0);
                        

                        double max = sqrt(pow((double)Broodwar->mapHeight()*TILE_SIZE,2)+pow((double)Broodwar->mapWidth()*TILE_SIZE,2));
                        double tmp = 1-2*(size/max);
                        double op;
                        if (tmp < 0) op = -1;
                        else op = 1;

                        double r = pow(tmp,2)*op;
                        double n = r + scout->ac->gamma * c - c_old; 
                        scout->ac->file << "tile " << r << " " << c << " " << c_old << " " << size << " " << n << endl;
                        
                        if (lmap) tiles[index] = pair<double,Position> (c,s);
                        else tiles[index] = pair<double,Position> (n,s);
                        index++;
                    }
                }
                if (lmap) std::sort(tiles.begin(), tiles.end(), sort_pair);
            }
            /* action */
            scout->move();
        }
      }
}



void ExampleAIModule::onSendText(std::string text) {
    if (text=="/analyze") {
        if (analyzed == false) {
            Broodwar->printf("Analyzing map... this may take a minute");
            CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)AnalyzeThread, NULL, 0, NULL);
        }
    } else if (text=="m") {
        isMap = (isMap +1) % 2;
    } else if (text=="g") {
        gui = (gui +1) % 2;
        Broodwar->setGUI((bool)gui);
    } else if (text=="l") {
        lmap = (lmap + 1) % 2;
    } else {
        Broodwar->printf("You typed '%s'!",text.c_str());
        Broodwar->sendText("%s",text.c_str());
  }
}

void ExampleAIModule::onReceiveText(BWAPI::Player* player, std::string text) {
  Broodwar->printf("%s said '%s'", player->getName().c_str(), text.c_str());
}

void ExampleAIModule::onPlayerLeft(BWAPI::Player* player) {
  Broodwar->sendText("%s left the game.",player->getName().c_str());
}

void ExampleAIModule::onNukeDetect(BWAPI::Position target) {}
void ExampleAIModule::onUnitDiscover(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitEvade(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitShow(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitHide(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitCreate(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitMorph(BWAPI::Unit* unit) {}
void ExampleAIModule::onUnitComplete(BWAPI::Unit *unit) {}
void ExampleAIModule::onUnitDestroy(BWAPI::Unit* unit) {
    if (unit->getType().isWorker()) {
        if (unit == scout->getScout()) scout->setScout(NULL);
    }
}

void ExampleAIModule::onUnitRenegade(BWAPI::Unit* unit) {
    if (unit->getType().isWorker()) {
        if (unit == scout->getScout()) scout->setScout(NULL);
    }
}
void ExampleAIModule::onSaveGame(std::string gameName) {
  Broodwar->printf("The game was saved to \"%s\".", gameName.c_str());
}

DWORD WINAPI AnalyzeThread() {
  BWTA::analyze();

  //self start location only available if the map has base locations
  if (BWTA::getStartLocation(BWAPI::Broodwar->self())!=NULL) {
    home       = BWTA::getStartLocation(BWAPI::Broodwar->self())->getRegion();
  }
  //enemy start location only available if Complete Map Information is enabled.
  if (BWTA::getStartLocation(BWAPI::Broodwar->enemy())!=NULL) {
    enemy_base = BWTA::getStartLocation(BWAPI::Broodwar->enemy())->getRegion();
  }
  analyzed = true;
  analysis_just_finished = true;
  return 0;
}