#include "MultiLayer.h"
#include <fstream>
#include <sstream>
#include <ctime>
#include <algorithm> 

#define PATH_READ "bwapi-data/write/"
#define PATH_WRITE "bwapi-data/write/"

MultiLayer::MultiLayer(int i, int o, int h, double a, string s) {
	inputs = i;
	outputs = o;
	neurons = h;
	alfa = a;
  file = s;

	w = new Matrix (neurons,inputs);
	v = new Matrix (outputs,neurons);	
	string tmp = PATH_READ;
  tmp.append(file);
  tmp.append("1.txt");
  if(!w->load(tmp)) {
    w->random(inputs);
  }
  tmp = PATH_READ;
  tmp.append(file);
  tmp.append("2.txt");
  if(!v->load(tmp)) {
    v->random(neurons);
  }
}

MultiLayer::~MultiLayer(void) {
	delete(w);
	delete(v);
}

void MultiLayer::save() {
  string tmp = PATH_WRITE;
  tmp.append(file);
  tmp.append("1.txt");
  w->save(tmp);
  tmp = PATH_WRITE;
  tmp.append(file);
  tmp.append("2.txt");
  v->save(tmp);
}

vector<double> MultiLayer::derivation(vector<double> e, vector<double> out) {
	vector<double> res;
	for (int i = 0; i < e.size(); i++) {
		res.push_back(e[i] * out[i]*(1-out[i]));
	}
	return res;
}

void MultiLayer::adjust(vector<double> error, Matrix *weight, vector<double> layer) {
	for (int i = 0; i < weight->h; i++) {
		for (int j = 0; j < weight->w; j++) {
			weight->set(j+i*weight->w,weight->get(j+i*weight->w) + alfa* error[i] * layer[j]);
		}
	}
}

void MultiLayer::learn(vector<double> error, vector<double> in) {
	Matrix *tran = v->transp();	
	vector<double> back = tran->multi(error,false);
	back = derivation(back,hidden);

	adjust(error,v,hidden);
	adjust(back,w,in);
	delete(tran);
}

vector<double> MultiLayer::run(vector<double> state) {
	hidden = w->multi(state,true);
	return v->multi(hidden,false);
}