#include "Scout.h"

using namespace BWAPI;

#define TRAINING 1
#define OFILE 1

#define DEF_EXPLO 120.0
#define ICOUNT 30
#define DEF_GAMMA 0.01
#define PATH_READ "bwapi-data/write/"
#define PATH_WRITE "bwapi-data/write/"

#define PI 3.14159265

Scout::Scout(BWAPI::Game* game, int range) {
    Broodwar = game;
    gamma = DEF_GAMMA;
    scout = NULL;
    canLearn = false;
    supervised = true;
    exploration = DEF_EXPLO;
    iterationCounter = 0;

    if (TRAINING) {
        string tmp = PATH_READ;
        tmp.append("LP_inc.txt");
        in.open(tmp.c_str());
        if (in != NULL) {
            in >> exploration >> iterationCounter;
            in.close();
        }
    }

    res.open("bwapi-data/write/LP_res.txt", ios::out | ios::app);
    res << endl;

    ac = new ActorCritic(21,gamma,exploration);
    //ac = new ActorCritic(3,gamma,exploration);
    pathfinder = new AStar(Broodwar);    

    srand(time(NULL));
}

Scout::~Scout() {
    delete(ac);
    delete(pathfinder);
}

void Scout::save() {
    ac->save(); 
    if (TRAINING) {
        string tmp = PATH_WRITE;
        tmp.append("LP_inc.txt");
        out.open(tmp.c_str());
        out << exploration << " " << iterationCounter;
        out.close();
    }
    res.close();
}


TilePosition Scout::getTilePosition(Position pos) {
    return TilePosition(floor(((double) pos.x() / TILE_SIZE) + 0.5),floor(((double) pos.y() / TILE_SIZE) + 0.5));
}
Unit* Scout::getScout() {
    return scout;
}
void Scout::setScout(Unit* unit) {
    scout = unit;
    if (unit != NULL) {
        hp = (double) unit->getHitPoints();
        shield = (double) unit->getShields();
    }
}
void Scout::setLocation(Position pos) {
    location = pos;
    path = pathfinder->find(scout->getPosition(),location);
}

Position Scout::getLocation() {
    return location;
}

bool Scout::reachedLocation(int frame) {
    if ((scout->getPosition().x() >= (location.x() - TILE_SIZE)) && (scout->getPosition().x() <= (location.x() + TILE_SIZE)) && (scout->getPosition().y() >= (location.y() - TILE_SIZE)) && (scout->getPosition().y() <= (location.y() + TILE_SIZE))) { 
        res << frame;
        return true;
    }   
    return false;
}

bool Scout::isOutOfMap() {
    if ((scout->getPosition().x() >= (Broodwar->mapWidth() - 2) * TILE_SIZE) || (scout->getPosition().y() >= (Broodwar->mapHeight() - 2) * TILE_SIZE) || (scout->getPosition().x() < (2 * TILE_SIZE)) || (scout->getPosition().y() < (2 * TILE_SIZE))) {
        res << -1;
        return true;
    }   
    return false;
}

Position Scout::generateLocation() {    
    //Position pos = Position((rand() % ((Broodwar->mapWidth() - 4 )* TILE_SIZE)) + 2 * TILE_SIZE,(rand() % ((Broodwar->mapHeight() - 4) * TILE_SIZE)) + 2 * TILE_SIZE);
    vector<double> op;
    op.push_back(-1);
    op.push_back(1);
    random_shuffle(op.begin(),op.end());

    int r = pow(500,2.0);    
    double rx = rand() % (r+1) ;
    double ry = r-rx;
    rx = sqrt(rx) * op[0];
    random_shuffle(op.begin(),op.end());
    ry = sqrt(ry) * op[0];  

    Position pos = Position((int) rx + scout->getPosition().x(), (int) ry + scout->getPosition().y());
    if ((pos.x() >= (Broodwar->mapWidth() - 2) * TILE_SIZE) || (pos.y() >= (Broodwar->mapHeight() - 2) * TILE_SIZE) || (pos.x() < (2 * TILE_SIZE)) || (pos.x() < (2 * TILE_SIZE))) return generateLocation();
      
    vector<Position> p = pathfinder->find(scout->getPosition(),pos); 
    if (p.size() > 0) return pos;
    else return generateLocation();
}

Position Scout::getNewLocation() {
    if (TRAINING) {
        exploration *= 0.999887654; 
        iterationCounter++;
    }
    return generateLocation();
    //return Position(100,100);
}


double Scout::pathSize(BWAPI::Position start, BWAPI::Position end, vector<BWAPI::Position> p) {
    double result = 0;
    if (p.size() > 2) {         
    	  result += sqrt(pow(p[1].x() - start.x(),2.0) + pow(p[1].y() - start.y(),2.0));

        for(int i = 1; i < p.size()-2; i++) {
            result += sqrt(pow(p[i+1].x() - p[i].x() ,2.0) + pow(p[i+1].y() - p[i].y() ,2.0));
        }
        result += sqrt(pow(p[p.size()-2].x() - end.x(),2.0) + pow(p[p.size()-2].y() - end.y(),2.0));   
    }
    else result += sqrt(pow(end.x() - start.x(),2.0) + pow(end.y() - start.y(),2.0));
    return result;
}

vector<double> Scout::createThreatMap() { 
    vector<double> map;
    int sight = scout->getType().sightRange();
    map.resize(16,0);
    Broodwar->drawCircleMap(scout->getPosition().x(), scout->getPosition().y(),sight,Colors::Orange,false);

    for(std::set<Unit*>::iterator u = Broodwar->getUnitsInRadius(scout->getPosition(),sight).begin();u!=Broodwar->getUnitsInRadius(scout->getPosition(),sight).end();u++) {
        if ((*u) != scout && (*u)->getPlayer()->isEnemy(Broodwar->self()) && (*u)->isCompleted()){        
            WeaponType weapon = (*u)->getType().groundWeapon();
            if (weapon != WeaponTypes::None) { 
                int weaponRange = weapon.maxRange();
                vector<double> positions;
                positions.push_back((*u)->getPosition().x());
                positions.push_back((*u)->getPosition().y());
                positions.push_back(positions[0]);                  positions.push_back(positions[1]);
                positions.push_back(positions[0] + weaponRange);    positions.push_back(positions[1]);
                positions.push_back(positions[0]);                  positions.push_back(positions[1] + weaponRange);
                positions.push_back(positions[0] - weaponRange);    positions.push_back(positions[1]);
                positions.push_back(positions[0]);                  positions.push_back(positions[1] - weaponRange);

                if (weaponRange > sight/4) {
                    positions.push_back(positions[0] + weaponRange/2);  positions.push_back(positions[1]);
                    positions.push_back(positions[0]);                  positions.push_back(positions[1] + weaponRange/2);
                    positions.push_back(positions[0] - weaponRange/2);  positions.push_back(positions[1]);
                    positions.push_back(positions[0]);                  positions.push_back(positions[1] - weaponRange/2);

                    positions.push_back(positions[0] + weaponRange/2);  positions.push_back(positions[1] + weaponRange/2);
                    positions.push_back(positions[0] + weaponRange/2);  positions.push_back(positions[1] - weaponRange/2);
                    positions.push_back(positions[0] - weaponRange/2);  positions.push_back(positions[1] + weaponRange/2);
                    positions.push_back(positions[0] - weaponRange/2);  positions.push_back(positions[1] - weaponRange/2);
                }
                //quadrant
                vector<bool> dirty;
                dirty.resize(16,false);
                for(int i = 0; i < positions.size(); i += 2) { 
                    double size = sqrt(pow(scout->getPosition().x() - positions[i],2.0) + pow(scout->getPosition().y() - positions[i+1],2.0));
                    double distance = pow(scout->getPosition().x() + size - positions[i],2.0) + pow(scout->getPosition().y() - positions[i+1],2.0);

                    double temp =(2 * pow(size,2.0) - distance) / (2 * pow(size,2.0));
                    if (temp < -1.0) temp = -1.0 ;
                    else if (temp > 1.0) temp = 1.0 ;

                    double angle = acos(temp) * 180.0 / PI;   
                    if (positions[i+1] > scout->getPosition().y()) angle = 360 - angle;

                    int quad = angle/45;        
                    if (size > sight/2) quad += 8;
                    Broodwar->drawCircleMap(positions[i], positions[i+1],1,Colors::Red,true);

                    if (!dirty[quad]) {
                        map[quad] += (double)(weapon.damageAmount() + weapon.damageBonus());
                        dirty[quad] = true;
                    }
                }
            }
        }
    }     
    return map;
}

vector<double> Scout::getState() { 
    vector<double> result;
    /* Shortest path */
    if (path.size() > 2) {
        result.push_back(path[1].x() - scout->getPosition().x());
        result.push_back(path[1].y() - scout->getPosition().y());      
    } else {
        result.push_back(location.x() - scout->getPosition().x());
        result.push_back(location.y() - scout->getPosition().y());
    }
    result = ac->normalizeVector(result);

    /* Path size */
    double dist = pathSize(scout->getPosition(),location, path);
    dist = 1/(1+dist*0.001);
    result.push_back(dist);

    /* HitPoints and Shields */
    result.push_back((double)(scout->getHitPoints())/(double)(scout->getType().maxHitPoints()));
    result.push_back((double)(scout->getShields())/(double)(scout->getType().maxShields()));

    /* ThreatMap */
    vector<double> map = createThreatMap();
    for (int i = 0; i < map.size(); i++) result.push_back(1/(1+map[i]*0.01));
    
    /* bias */
    result.push_back(-1);
    return result;
}

vector<double> Scout::getOtherState(Position unit, double & dist) { 
    vector<double> result;
    // ShortestPath
    vector<Position> p = pathfinder->find(unit,location);
    if (p.size() > 2) {
        result.push_back(p[1].x() - unit.x());
        result.push_back(p[1].y() - unit.y());   
        result = ac->normalizeVector(result);
        dist = pathSize(unit,location,p);
        result.push_back(1/(1+dist*0.001));
    } else {
        ac->file << "nopath " <<  unit.x()<< " " << unit.y() << "   " <<  unit.x()/TILE_SIZE << " " << unit.y()/TILE_SIZE <<  endl;
        result.push_back(location.x() - unit.x());
        result.push_back(location.y() - unit.y());
        result = ac->normalizeVector(result);
        dist = pathSize(unit,location,p);
        result.push_back(1/(1+dist*0.01));
    }
    /* HitPoints and Shields */
    result.push_back(scout->getHitPoints()/scout->getType().maxHitPoints());
    result.push_back(scout->getShields()/scout->getType().maxShields());

    /* ThreatMap */
    vector<double> map = createThreatMap();
    for (int i = 0; i < map.size(); i++) result.push_back(1/(1+map[i]*0.01)); 

    /* bias */
    result.push_back(-1);
    return result;
}


double Scout::getReward(Position pos) { 
    double result = 0;
    /* scout is alive */
    if (scout->exists()) {
        /* path size */
        double size = pathSize(pos,location,path);     
        double max = sqrt(pow((double)Broodwar->mapHeight() * TILE_SIZE, 2) + pow((double)Broodwar->mapWidth() * TILE_SIZE, 2));
        double tmp = 1-2*(size/max);
        double op;

        if (tmp < 0) op = -1;
        else op = 1;
        result += pow(tmp,2)*op;    

        /* scout was damaged */
        double currrentHP = scout->getHitPoints();
        double currentShield = scout->getShields();

        if (currrentHP < hp || currentShield < shield) {
            ac->file << "Damage: " << result << " ";
            result -= (hp - currrentHP)/scout->getType().maxHitPoints() * 0.01;
            result -= (shield - currentShield)/scout->getType().maxShields()* 0.001;
            ac->file << result << endl;
            hp = currrentHP;
            shield = currentShield;
        }

    } else {
        result -= hp/scout->getType().maxHitPoints() * 0.01;
        result -= shield/scout->getType().maxShields()* 0.001;
    }
    return result;        
}

Position Scout::getPosition(vector<double> action) {
    return Position (scout->getPosition().x() + (int)(action[0] * TILE_SIZE),scout->getPosition().y() + (int)(action[1] * TILE_SIZE));
}

void Scout::move() {
    if (scout != NULL) {
        /* generate state s */
        state = getState();

        /* choose action a and explore it */
        action = ac->run(state);
        vector<double> normalized = ac->normalizeVector(action);
        vector<double> explored = ac->explore(normalized);

        if (OFILE) ac->file << "Normalized: " << normalized[0] << " " << normalized[1] << endl;

        /* draw vectors */
        drawVector.resize(0);
        drawVector.push_back(action[0]);
        drawVector.push_back(action[1]);
        drawVector.push_back(explored[0]);
        drawVector.push_back(explored[1]);

        /* compute error */
        error.resize(0);
        if (TRAINING && iterationCounter < ICOUNT && getReward(getPosition(normalized)) > getReward(getPosition(explored))) {
            /*supervised learning */
            supervised = true;
            for (int i = 0; i < action.size(); i++) error.push_back(state[i] - action[i]);
        } else {
            supervised = false;
            for (int i = 0; i < action.size(); i++) error.push_back(explored[i] - action[i]);   
        }
       
        /* do explored action a */
        Position p = getPosition(explored);
        if (Broodwar->mapWidth()*TILE_SIZE > p.x() || Broodwar->mapHeight()*TILE_SIZE > p.y() || p.x() >= 0 || p.y() >= 0 ) {
            scout->move(p);
        }
        canLearn = true;
    }
}

void Scout::learn() {
    /* begin of state s */
    if (canLearn && scout != NULL) {  
        /* learn from state s-1 */
        path = Scout::pathfinder->find(scout->getPosition(),location);
        double reward = getReward(scout->getPosition());      

        if (scout->exists()) {
            /* scout is alive */
            ac->learn(reward,state,getState(),error, supervised);
        } else {
            /* scout was destroyed */
            vector<double> empty;
            ac->learn(reward,state,empty,error,supervised);
            scout = NULL;
        }
        canLearn = false;
    }
}